from django import forms


class MoneyWidget(forms.MultiWidget):
    def __init__(self, amount_widget=forms.TextInput, currency_widget=forms.Select, choices=None, attrs=None):
        if isinstance(currency_widget, type):
            currency_widget = currency_widget(choices=choices)

        super().__init__((amount_widget, currency_widget), attrs)

    def decompress(self, value):
        if value is not None:
            if isinstance(value, (list, tuple)):
                return value
            return [value.amount, value.currency]
        return None



class MoneyException(Exception):
    pass


class CurrencyException(MoneyException):
    pass


class InvalidCurrencyFormat(CurrencyException, ValueError):
    def __init__(self, currency):
        msg = (
            'Currency not in ISO 4217 format: \'{}\'.'
        ).format(currency)
        super().__init__(msg)


class ExchangeError(MoneyException):
    pass


class InvalidExchangeBackend(ExchangeError):
    def __init__(self):
        msg = (
            'Parameter \'backend\' isn\'t a subclass of money.exchange.BaseBackend.'
        )
        super().__init__(msg)


class ExchangeBackendNotSet(ExchangeError):
    def __init__(self):
        msg = (
            'Exchange backend hasn\'t been set yet. Set it using '
            'money.xrates.backend = \'path.to.backend\'.'
        )
        super().__init__(msg)


class ExchangeRateNotFound(ExchangeError):
    def __init__(self, backend, origin, target):
        msg = (
            'Rate not found in backend \'{}\': {}/{}.'
        ).format(backend, origin, target)
        super().__init__(msg)



from django import forms

from money import Money
from money.django.widgets import MoneyWidget


class MoneyField(forms.MultiValueField):
    def __init__(self, amount_widget=None, default_amount=None, max_value=None, min_value=None, max_digits=None,
                 decimal_places=None, currency_widget=None, default_currency=None, choices=(), *args, **kwargs):
        amount_field = forms.DecimalField(
            widget=amount_widget or forms.DecimalField.widget,
            max_value=max_value, min_value=min_value, max_digits=max_digits, decimal_places=decimal_places
        )
        currency_field = forms.ChoiceField(
            widget=currency_widget or forms.ChoiceField.widget,
            choices=choices
        )

        self.widget = MoneyWidget(
            amount_widget=amount_field.widget,
            currency_widget=currency_field.widget,
            choices=choices
        )

        super().__init__((amount_field, currency_field), *args, **kwargs)

        self.initial = [default_amount, default_currency]

    def compress(self, data_list):
        if data_list:
            if not self.required and (data_list[0] in self.empty_values or data_list[1] in self.empty_values):
                return None
            return Money(*data_list[:2])
        return None



from decimal import Decimal

from django.db import models
from django.utils.translation import gettext_lazy as _

from money import Currency, Money
from money.django import forms


def _currency_field_name(name): return '{}_currency'.format(name)


class MoneyFieldProxy(object):
    def __init__(self, field):
        self.field = field
        self.currency_name = _currency_field_name(self.field.name)

    def __set__(self, obj, value):
        if isinstance(value, Money):
            obj.__dict__[self.field.name] = value
            setattr(obj, self.currency_name, value.currency)
        else:
            obj.__dict__[self.field.name] = self.field.to_python(value)

    def __get__(self, obj, type=None):
        if obj is None:
            return self

        amount = obj.__dict__[self.field.name]
        if amount is None:
            return None
        else:
            return Money(amount, getattr(obj, self.currency_name))


class CurrencyField(models.CharField):
    description = _("Currency Object")

    def __init__(self, *args, **kwargs):
        default = kwargs.get('default', None)
        if isinstance(default, Currency):
            default = default.code

        kwargs['max_length'] = 3
        kwargs['default'] = default

        super().__init__(*args, **kwargs)


class MoneyField(models.DecimalField):
    description = _("Money Object")

    def __init__(self, verbose_name=None, name=None, max_digits=None, decimal_places=None, default_currency=None,
                 currency_choices=None, **kwargs):
        default = kwargs.get('default', None)
        if not default_currency and default is not None:
            default_currency = default.currency

        self.default_currency = default_currency
        self.currency_choices = currency_choices
        self.nullable = kwargs.get('null', False)

        super().__init__(verbose_name, name, max_digits, decimal_places, **kwargs)

    def contribute_to_class(self, cls, name, private_only=False):
        currency_field = CurrencyField(
            default=self.default_currency, choices=self.currency_choices,
            null=self.nullable, editable=False
        )
        currency_field.creation_counter = self.creation_counter
        cls.add_to_class(_currency_field_name(name), currency_field)

        super().contribute_to_class(cls, name, private_only)

        setattr(cls, self.name, MoneyFieldProxy(self))

    def deconstruct(self):
        name, path, args, kwargs = super().deconstruct()

        if isinstance(self.default, Money):
            kwargs['default'] = Decimal.__str__(self.default)
        if self.default_currency is not None:
            kwargs['default_currency'] = str(self.default_currency)
        if self.currency_choices is not None:
            kwargs['currency_choices'] = self.currency_choices

        return name, path, args, kwargs

    def formfield(self, **kwargs):
        default = self.default
        if isinstance(default, Money):
            default = default.amount

        return super().formfield(**{
            'form_class': forms.MoneyField,
            'default_amount': default,
            'default_currency': self.default_currency,
            'choices': self.currency_choices,
            **kwargs
        })



from decimal import Decimal, ROUND_HALF_UP

from babel.numbers import format_currency

from money import Currency, xrates
from money.exceptions import ExchangeBackendNotSet, ExchangeRateNotFound


def _make_comparison_operator(name):
    method = getattr(Decimal, name, None)

    def operator_func(self, other, *args):
        if isinstance(other, Money):
            other = other.to(self._currency)

        return method(self, other, *args)

    return operator_func


def _make_arithmetic_operator(name):
    method = getattr(Decimal, name, None)

    def operator_func(self, other, *args):
        if isinstance(other, Money):
            other = other.to(self._currency)

        result = method(self, other, *args)
        if result is NotImplemented:
            return NotImplemented

        return self.__class__(result, self._currency)

    return operator_func


class Money(Decimal):
    """Class representing a monetary amount."""

    _rounding_mode = ROUND_HALF_UP

    __slots__ = ('_currency',)

    def __new__(cls, amount, currency):
        self = super().__new__(cls, amount)

        if not isinstance(currency, Currency):
            currency = Currency(str(currency))

        self._currency = currency

        return self

    @property
    def amount(self):
        """Returns the amount rounded to the correct number of decimal places for the currency."""

        decimal_precision = Decimal(str(1 / (10 ** self._currency.precision)).rstrip('0'))
        return self.quantize(decimal_precision, rounding=Money._rounding_mode)

    @property
    def currency(self):
        """Returns the currency."""

        return self._currency

    def __repr__(self):
        return 'Money({}, \'{}\')'.format(super().__repr__(), self._currency)

    def __str__(self):
        return self.format()

    def __reduce__(self):
        return self.__class__, (Decimal.__str__(self), self._currency)

    def __eq__(self, other):
        result = super().__eq__(other)
        if result is NotImplemented:
            return NotImplemented

        if isinstance(other, Money):
            return result and other._currency == self._currency
        else:
            return result

    def __ne__(self, other):
        return not self == other

    __hash__ = Decimal.__hash__

    __lt__ = _make_comparison_operator('__lt__')
    __le__ = _make_comparison_operator('__le__')
    __gt__ = _make_comparison_operator('__gt__')
    __ge__ = _make_comparison_operator('__ge__')

    __add__ = _make_arithmetic_operator('__add__')
    __radd__ = _make_arithmetic_operator('__radd__')
    __sub__ = _make_arithmetic_operator('__sub__')
    __rsub__ = _make_arithmetic_operator('__rsub__')
    __mul__ = _make_arithmetic_operator('__mul__')
    __rmul__ = _make_arithmetic_operator('__rmul__')
    __truediv__ = _make_arithmetic_operator('__truediv__')
    __rtruediv__ = _make_arithmetic_operator('__rtruediv__')
    __divmod__ = _make_arithmetic_operator('__divmod__')
    __rdivmod__ = _make_arithmetic_operator('__rdivmod__')
    __mod__ = _make_arithmetic_operator('__mod__')
    __rmod__ = _make_arithmetic_operator('__rmod__')
    __floordiv__ = _make_arithmetic_operator('__floordiv__')
    __rfloordiv__ = _make_arithmetic_operator('__rfloordiv__')
    __pow__ = _make_arithmetic_operator('__pow__')
    __rpow__ = _make_arithmetic_operator('__rpow__')

    def __neg__(self):
        return self.__class__(super().__neg__(), self._currency)

    def __pos__(self):
        return self.__class__(super().__pos__(), self._currency)

    def __abs__(self):
        return self.__class__(super().__abs__(), self._currency)

    def __round__(self, n=0):
        return self.__class__(super().__round__(n), self._currency)

    def __floor__(self):
        return self.__class__(super().__floor__(), self._currency)

    def __ceil__(self):
        return self.__class__(super().__ceil__(), self._currency)

    def to(self, currency):
        """Returns the equivalent money object in another currency."""

        if currency == self._currency:
            return self

        if xrates.backend is None:
            raise ExchangeBackendNotSet()

        if not isinstance(currency, Currency):
            currency = Currency(str(currency))

        rate = xrates.backend.quotation(self._currency.code, currency.code)
        if rate is None:
            raise ExchangeRateNotFound(xrates.backend_name, self._currency, currency)

        return self.__class__(self * rate, currency)

    def format(self, locale='en_US'):
        """Returns a string of the currency formatted for the specified locale."""

        return format_currency(self, self.currency.code, locale=locale).replace(u'\xa0', u' ')

    @classmethod
    def set_rounding_mode(cls, mode):
        cls._rounding_mode = mode
